/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;

import org.junit.runner.JUnitCore;

import com.ibm.rqm.adapter.library.data.AdapterConstants;

/**
 * @author swijenay
 * @author tneal
 *
 * Launch the specified JUnit testcases/testsuites, logging their pass/fail results
 * in a report file.
 */
public class JUnitLauncher{

	public static final String COMMAND_LINE_ARGUMENT_TEST = "-test";
	public static final String COMMAND_LINE_ARGUMENT_REPORT_FILE = "-reportFile";
	public static final String COMMAND_LINE_ARGUMENT_PLATFORM = "-platform";
	public static final String COMMAND_LINE_ARGUMENT_DESCRIPTION = "-description";
	public static final String COMMAND_LINE_ARGUMENT_SCREENSHOT_DIR = "-screenshotDir";
	public static final String COMMAND_LINE_ARGUMENT_PROPERTIES = "-properties";
	public static final String DEFAULT_REPORT_NAME = "JUnitReport.ser";
	public static final String COMMAND_LINE_ARGUMENT_RESULTATTACH_FILE = AdapterConstants.RESULT_ATTACHMENT_FILE_PROPERTY;

	public static void main(String[] arguments){
		try{
			String testClassNameList = null;
			String reportFile = null;
			String platform = null;
			String description = null;
			File screenshotDirectory = null;
			Hashtable<String,String> properties = new Hashtable<String, String>();

			for(int i = 0; i < arguments.length; i++){
				if((arguments[i].compareToIgnoreCase(COMMAND_LINE_ARGUMENT_REPORT_FILE) == 0) && 
						(i + 1 < arguments.length)){
					reportFile = arguments[i+1].trim();
				}
				// COMMAND_LINE_ARGUMENT_PLATFORM not currently used by JUnit Selenium adapter, reserved for future use
				else if((arguments[i].compareToIgnoreCase(COMMAND_LINE_ARGUMENT_PLATFORM) == 0) && 
						(i + 1 < arguments.length)){
					platform = arguments[i+1].trim();
				}
				else if((arguments[i].compareToIgnoreCase(COMMAND_LINE_ARGUMENT_TEST) == 0) && 
						(i + 1 < arguments.length)){
					testClassNameList = arguments[i+1].trim();
				}
				// COMMAND_LINE_ARGUMENT_DESCRIPTION not currently used by JUnit Selenium adapter, reserved for future use
				else if((arguments[i].compareToIgnoreCase(COMMAND_LINE_ARGUMENT_DESCRIPTION) == 0) && 
						(i + 1 < arguments.length)){
					description = arguments[i+1].trim();
				}
				else if((arguments[i].compareToIgnoreCase(COMMAND_LINE_ARGUMENT_SCREENSHOT_DIR) == 0) && 
						(i + 1 < arguments.length)){
					screenshotDirectory = new File(arguments[i+1].trim());
				}
				else if((arguments[i].compareToIgnoreCase(COMMAND_LINE_ARGUMENT_PROPERTIES) == 0) && 
						(i + 1 < arguments.length)){
					addProperties(properties, arguments[i+1].trim());
				}
				else if((arguments[i].indexOf(COMMAND_LINE_ARGUMENT_RESULTATTACH_FILE) == 0) ){
					addProperties(properties, arguments[i].trim());
				}
			}
			
			if(testClassNameList != null){
				File reportFileInstance;
				
				if(reportFile == null){
					reportFileInstance = new File(new File(System.getProperty("user.home")), DEFAULT_REPORT_NAME);
				}
				else{
					reportFileInstance = new File(reportFile);
				}
				
				if(platform == null){
					platform = MessageFormat.format("{0} {1}", new Object[]{System.getProperty("os.name"), System.getProperty("os.version")});
				}

				if(description == null){
					description = "JUnit Test Results";
				}
				
				
				if (screenshotDirectory == null) {
					screenshotDirectory = new File(reportFileInstance.getParentFile(), "imgs_" + System.currentTimeMillis());
				}
				if (!screenshotDirectory.exists())
					screenshotDirectory.mkdirs();
				properties.put("selenium.screenshot.dir", screenshotDirectory.getAbsolutePath());
								
				final JUnitReport report = new JUnitReport(description, platform);

				Enumeration<String> propertyNames = properties.keys();
				while(propertyNames.hasMoreElements()){
					String propertyName = propertyNames.nextElement();
					System.setProperty(propertyName, properties.get(propertyName));
				}
				
				JUnitCore jUnitCore = new JUnitCore();
				jUnitCore.addListener(new JUnitListener(report, screenshotDirectory));				
				String[] testClassNames = testClassNameList.split(",");

				for(String testClassName : testClassNames){
					Class<?> testClass = JUnitLauncher.class.getClassLoader().loadClass(testClassName);
					jUnitCore.run(testClass);
				}
				
				// Serialize the report object so it can be read by the JUnit Selenium adapter and it's contents added to the execution result.
				FileOutputStream fos = new FileOutputStream(reportFileInstance);
				ObjectOutputStream oos = new ObjectOutputStream(fos);
				oos.writeObject(report);
				oos.close();
				fos.close();
			}
			else{
				System.err.println("Invalid Execution!"); 
				System.err.println("Usage: -test <comma separated list of test suites or test cases> [-reportFile <path to report file>] [-platform <platform description>] [-description <test result description for report>");
				System.exit(1);
			}
		}
		catch(Exception e){
			e.printStackTrace();
			System.exit(1);
		}
		
		// Call exit to cleanup any remaining threads that might have been started by the JUnit test(s)
		System.exit(0);
	}

	/**
	 * Parse the name/value pairs given in input and add them to the hash table.
	 * The format of the input string is name=value[,name=value]...
	 * 
	 * @param properties
	 * @param input
	 */
	private static void addProperties(Hashtable<String, String> properties, String input) {
		if (input != null) {
			// Split on comma, preserving any commas in a single-quoted value and remove leading/trailing quotes
			String[] pairs = input.split(",(?=([^\']*\'[^\']*\')*[^\']*$)");
			for (String pair : pairs) {
				int equalNdx = -1;
				if ((equalNdx = pair.indexOf('=')) <= 0)
					continue;
				String value = null;
				String name = pair.substring(0, equalNdx).trim();
				if (equalNdx < pair.length() - 1) {
					value = pair.substring(equalNdx+1).trim();
					// Remove leading and trailing quotes if present from quoted value
					if (value.startsWith("\'") && value.endsWith("\'")) {
						value = value.substring(1, value.length() - 1);
					}
				}
				if (name.length() > 0 && value != null && value.length() > 0)
					properties.put(name, value);
			}
		}
	}
}
