/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Properties;
import java.util.Vector;

import org.junit.runner.Description;
import org.junit.runner.Result;
import org.junit.runner.notification.Failure;
import org.junit.runner.notification.RunListener;

/**
 * @author swijenay
 * @author tneal
 * 
 *         JUnit listener for logging all events related to a JUnit test run.
 *         Logged events will be stored in a JUnitReport object describing the
 *         results of a testsuite/testcase run. The JUnitReport object will be
 *         serialized and later used to add information to a execution result.
 * 
 */
public class JUnitListener extends RunListener {
	JUnitReport report = null;
	File screenshotDirectory = null;
	boolean errorOrFailureReported = false;
	HashSet<File> claimedScreenshots = new HashSet<File>();
	HashMap<String, Long> startTimeMap = new HashMap<String, Long>();
	int numTests = 0;
	int testsRun = 0;

	public JUnitListener(JUnitReport report, File screenShotDirectory) {
		this.report = report;
		this.screenshotDirectory = screenShotDirectory;
	}
	
	public void testRunStarted(Description description) throws Exception {
		numTests = description.testCount();
	}

	public void testRunFinished(Result result) throws Exception {
	}
	
	public void testStarted(Description description) throws Exception {
		this.startTimeMap.put(description.getDisplayName(), System.currentTimeMillis());
		errorOrFailureReported = false;	
		// Do a progress update here so the test name will be set correctly
		updateProgress(description.getDisplayName(), (int)(((double)testsRun/(double)numTests)*100));
	}

	public void testFinished(Description description) throws Exception {
		if(!errorOrFailureReported){
			report.getTestCaseReporter(getTestCaseName(description.getDisplayName()))
				.report(getTestMethodName(description.getDisplayName()), 
						null, getUnclaimedScreenshots(),
						JUnitReportEntryType.PASS, startTimeMap.get(description.getDisplayName()));
		}
		// Do a progress update here so the percent done will be set correctly
		testsRun++;
		updateProgress(description.getDisplayName(), (int)(((double)testsRun/(double)numTests)*100));
	}

    public void testFailure(Failure failure) {
		errorOrFailureReported = true;
		report.getTestCaseReporter(getTestCaseName(failure.getDescription().getDisplayName()))
			.report(getTestMethodName(failure.getDescription().getDisplayName()), 
					getStackTrace(failure.getException()), getUnclaimedScreenshots(),
					JUnitReportEntryType.FAIL, startTimeMap.get(failure.getDescription().getDisplayName()));
	}

	/**
	 * Return the name of the JUnit test case. If we encounter errors trying to
	 * parse the test case name from displayName, we will return the entire
	 * displayName.
	 */
	private String getTestCaseName(String displayName) {
		try {
			int startOfTestCaseName = displayName.indexOf("(") + 1;
			int endOfTestCaseName = displayName.indexOf(")", startOfTestCaseName);

			return displayName.substring(startOfTestCaseName, endOfTestCaseName);
		} catch (IndexOutOfBoundsException e) {
			return displayName;
		}
	}

	/**
	 * Return the name of the test method within a JUnit test case. If we
	 * encounter errors trying to parse the test method name from displayName,
	 * we will return the entire displayName.
	 */
	private String getTestMethodName(String displayName) {
		try {
			int startOfTestCaseName = displayName.indexOf("(") + 1;
			int endOfTestCaseName = displayName.indexOf(")", startOfTestCaseName);

			return displayName.substring(0, startOfTestCaseName)
					+ displayName.substring(endOfTestCaseName);
		} catch (IndexOutOfBoundsException e) {
			return displayName;
		}
	}

	/**
	 * Search through the screenshots directory looking for any image files that
	 * have not already been "claimed" (saved into the claimedScreenshots map).
	 * If any "unclaimed" screenshots are found they will be returned by this
	 * method and also added to the "claimed" map so subsequent calls will not
	 * include them.
	 */
	private Vector<File> getUnclaimedScreenshots() {
		Vector<File> unclaimedScreenshots = new Vector<File>();
		
		File[] screenshots = screenshotDirectory.listFiles(new FileFilter() {
			public boolean accept(File pathname) {
				return pathname.getName().toLowerCase().endsWith(".png") ||
					   pathname.getName().toLowerCase().endsWith(".jpg") ||
					   pathname.getName().toLowerCase().endsWith(".jpeg") ||
					   pathname.getName().toLowerCase().endsWith(".gif") ||
					   pathname.getName().toLowerCase().endsWith(".tif") ||
					   pathname.getName().toLowerCase().endsWith(".bmp");
			}
		});
		
		for (File screenshot : screenshots){
			if(!claimedScreenshots.contains(screenshot)){
				unclaimedScreenshots.add(screenshot);
				claimedScreenshots.add(screenshot);
			}
		}
		
		return unclaimedScreenshots;
	}
	
	/**
	 * Update the test progress with the name of the JUnit test being executed
	 */
	private void updateProgress(String testName, int percentDone) {
		// The last test will set percentDone == 100, make that 99 or progress update won't occur
		if (percentDone == 100)
			percentDone = 99;
		String path = System.getenv("qm_ExecutionPropertiesFile"); //$NON-NLS-1$
		if (path != null) {
			try {
				FileOutputStream progressStream = new FileOutputStream(new File(path));
				Properties progressProps = new Properties();
				progressProps.put("progress", String.valueOf(percentDone)); //$NON-NLS-1$
				progressProps.put("comment", "Executing test: " + testName); //$NON-NLS-1$
				progressProps.store(progressStream, ""); //$NON-NLS-1$
				progressStream.close();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} 
		}
		return;
	}

	/*
	 * Return a String representation of a exception traceback.
	 */
	private String getStackTrace(Throwable t) {
		if (t == null)
			return null;
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		t.printStackTrace(pw);
		pw.flush();
		sw.flush();
		return sw.toString();
	}
}
