/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Semaphore;

import org.apache.abdera.model.Element;

import com.ibm.rqm.adapter.library.connector.AbstractTestToolAdapter;
import com.ibm.rqm.adapter.library.data.AbstractProcessor;
import com.ibm.rqm.adapter.library.data.AdapterConstants;
import com.ibm.rqm.adapter.library.data.AdapterException;
import com.ibm.rqm.adapter.library.data.ILogResult;
import com.ibm.rqm.adapter.library.data.NewRequester;


/**
 * The processor which takes in tasks and instructions fetched by the requester
 * from the server and handles it appropriately to the adapter specific
 * implementations of the {@link AbstractTestToolAdapter}.
 * 
 * @author ddoraswa
 * 
 */
public class MultiThreadQMCommandLineProcessor extends AbstractProcessor implements AdapterConstants {

	private ILogResult logResult = null;
	private List <MultiThreadCommandLineExecutionThread> cmdExeThreads = new ArrayList <MultiThreadCommandLineExecutionThread>();
	private int MAX_AVAILABLE = 1;
	private Semaphore available = null;
	
	public MultiThreadQMCommandLineProcessor(NewRequester requester, int maxServedRequests) {
		super(requester);
		available = new Semaphore(maxServedRequests, true);
		MAX_AVAILABLE = maxServedRequests;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see com.ibm.rqm.adapter.library.data.AbstractProcessor#processInstruction(org.apache.abdera.model.Element)
	 */
	@Override
	protected synchronized void processInstruction(Element instruction) {
		String type = instruction.getAttributeValue(INSTRUCTION_TYPE);
		String task = instruction.getAttributeValue(INSTRUCTION_TASK_ID);
		
		if (type.equalsIgnoreCase("CANCEL") && cmdExeThreads != null) {
			int index = 0;
				for(MultiThreadCommandLineExecutionThread th:cmdExeThreads) {
					if(th.getTaskId().equals(task)) {
						th.getCommandLineAdapter().stopExecution();
						break;
					}
					index++;
				}
				cmdExeThreads.remove(index);
				if(cmdExeThreads.size() < MAX_AVAILABLE) {
					available.release();
					setBusy(false);
				}
		}

	}

	
	protected synchronized void processExecution(String url, Element element){
		try {
			available.acquire();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		MultiThreadCommandLineExecutionThread cmdExeThread = new MultiThreadCommandLineExecutionThread(url, element, requester, this);
		cmdExeThreads.add(cmdExeThread);
		cmdExeThread.start();
		setBusy(true);
		
	}
	/*
	 * (non-Javadoc)
	 * 
	 * @see com.ibm.rqm.adapter.library.AbstractProcessor#processImport(java.lang.String,
	 *      java.util.Map, com.ibm.rqm.adapter.library.Requester)
	 */
	@Override
	protected void processImport(String url, Element element) {

		// Get most of the data from the propertiesMap
		String location = requester.getValue("qmtask:scriptPath", element);
		File folder = new File(location);
		if(folder.exists())
		{	
			if(folder.isFile())
			{
				try {
					requester.uploadResources(url, (new String[] {folder.getName()}),folder.getParentFile().getPath());
				} catch (AdapterException e) {
					Logger.Log.error("Could not upload the file: " + e.getMessage(), e);
				}
			}
			else
			{
				File[] files = folder.listFiles(new FileFilter() {
					public boolean accept(File pathname) {
						return (pathname.isFile());
					}
				});
				List<String> resources = new ArrayList<String>(files.length);
				for (int i = 0; i < files.length; i++) {
					resources.add(files[i].getName());
				}
				try {
					requester.uploadResources(url, resources.toArray(new String[] {}),folder.getPath());
				} catch (AdapterException e) {
					Logger.Log.error("Could not upload resources: " + e.getMessage(), e);
				}
			}
		}
		else
		{
			try {
				requester.uploadResources(url, (new String[] {}), null);
			} catch (AdapterException e) {
				Logger.Log.error("Could not upload empty array: " + e.getMessage(), e);
			}
		}
	}
	public synchronized void setBusy(boolean busy){
		if(cmdExeThreads.size() >= MAX_AVAILABLE) {
			super.setBusy(true);
		} else {
			super.setBusy(false);
		}
	}
	public synchronized void removeThread(MultiThreadCommandLineExecutionThread thread){
		int index = 0;
		for(MultiThreadCommandLineExecutionThread th:cmdExeThreads) {
			if(th.getTaskId().equals(thread.getTaskId())) {
				cmdExeThreads.remove(index);
				break;
			}
			index++;
		}
		if(cmdExeThreads.size() < MAX_AVAILABLE) {
			available.release();
		}
		setBusy(false);
		
	}
}
