/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2014. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.rqm.adapter.commandline;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.InvalidPropertiesFormatException;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import com.ibm.rqm.adapter.library.CommandLineParser;
import com.ibm.rqm.adapter.library.ConnectionInfo;
import com.ibm.rqm.adapter.library.connection.AdapterEncryption;
import com.ibm.rqm.adapter.library.connection.AesEncryptUtil;
import com.ibm.rqm.adapter.library.connection.Base64;
import com.ibm.rqm.adapter.library.data.AbstractProcessor;
import com.ibm.rqm.adapter.library.data.AdapterConstants;
import com.ibm.rqm.adapter.library.data.AdapterInformation;
import com.ibm.rqm.adapter.library.data.NewRequester;
import com.ibm.rqm.adapter.library.data.RQMAdapter;
import com.ibm.rqm.adapter.library.data.ServerInformation;

/**
 * The entry point to the adapter. This is the class which is instantiated to
 * start up the adapter. It keeps all the information about the adapter instance
 * and delegates the task of communicating to the QM server via
 * {@link NewRequester} and the adapter specific implementation of
 * {@link AbstractProcessor} (for Command Line adapter its
 * {@link QMCommandLineProcessor})
 * 
 * <br>
 * Once the adapter is instantiated, it starts polling the server for new
 * instructions via NewRequester {@link RQMAdapter#start()}
 * 
 * If an instruction is fetched the adapter then delegates the servicing of the
 * instruction to the processor {@link QMCommandLineProcessor}.
 * 
 * @author amathur
 * 
 */
public class QMCommandLineAdapter extends RQMAdapter {

	/**
	 * <code>ADAPTER_TYPE</code> The type which is registered with the QM
	 * server via extension point
	 * <code>com.ibm.rqm.planning.common.executionScriptType</code> attribute
	 * <code>id</code>
	 */
	protected static final String CLA_ADAPTER_TYPE = "com.ibm.rqm.adapter.commandline"; //$NON-NLS-1$
	private static final String CLA_ADAPTER_DESCRIPTION = "Its a command Line Adapter";
	private static final String CLA_ADAPTER_DEFAULT_NAME = "Commandline Adapter";

	protected static final String SELENIUM_ADAPTER_TYPE = "com.ibm.rqm.adapter.selenium"; //$NON-NLS-1$
	private static final String SELENIUM_ADAPTER_DESCRIPTION = "JUnit Selenium Adapter";
	private static final String SELENIUM_ADAPTER_DEFAULT_NAME = "JUnit Selenium Adapter";

	private static final String COMMENT = "RQM Command Line Configuration File"; //$NON-NLS-1$
	protected static final String DEFAULT_PREFERENCES_NAME = "commandline.properties";
	protected static final String SELENIUM_PREFERENCES_NAME = "selenium.properties";
	private static final String[] ADAPTER_CAPABILITIES = { AdapterInformation.IMPORT_CAPABILITY, AdapterInformation.EXECUTE_CAPABILITY }; // , AdapterInformation.RECORD_CAPABILITY};
	
	/**
	 * @param connectionInfo
	 * @param adapterType
	 * @param adapterName
	 * @param sleepTime
	 * @param verbose
	 */
	public QMCommandLineAdapter(ServerInformation server, AdapterInformation adapter) {
		super(server, adapter);
		
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see com.ibm.rqm.adapter.library.data.RQMAdapter#createProcessor(com.ibm.rqm.adapter.library.data.NewRequester)
	 */
	protected AbstractProcessor createProcessor(NewRequester requester) {
		int maxServedRequests = 1;
		try{
			maxServedRequests = CommandLinePreferences.prefs().getIntProperty(CommandLinePreferences.CLA_MAX_SERVED_REQUESTS, -1);
		}catch(NumberFormatException nfe){
			Logger.Log.info(CommandLinePreferences.CLA_MAX_SERVED_REQUESTS + " entry is bad in preference file, the adapter will serve single request at a time.");
		}
		
		if(requester.getAdapter().getType().equals(SELENIUM_ADAPTER_TYPE) && maxServedRequests > 1){
			//If this is the selenium adapter and the max thread is set in selenium.properties file.
			//Fail to start and ask for updation in the properties file.
			Logger.Log.error(CommandLinePreferences.CLA_MAX_SERVED_REQUESTS + " entry is alllowed only for " + CLA_ADAPTER_TYPE + " in " + DEFAULT_PREFERENCES_NAME + ". Remove the entry " + CommandLinePreferences.CLA_MAX_SERVED_REQUESTS + " and then start Selenium adapter");
			System.exit(1);
		}
		
		if(maxServedRequests > 1) {
			//Return the multi-threaded version of the adapter
			if (Logger.Log.isDebugEnabled()) {
				Logger.Log.debug("Max served requests found to be " + maxServedRequests + ", starting adapter in multi threaded mode");
			}
			return new MultiThreadQMCommandLineProcessor(requester, maxServedRequests);
		} else {
			//return the single threaded version of the adapter.
			if (Logger.Log.isDebugEnabled()) {
				Logger.Log.debug("Max served requests found to be " + maxServedRequests + ", starting adapter in single threaded mode");
			}
			return new QMCommandLineProcessor(requester);
		}
	}

	/**
	 * The supported arguments are
	 * {@link CommandLineParser#ADAPTER_ID_ARGUMENT_NAME},
	 * {@link CommandLineParser#PASSWORD_ARGUMENT_NAME},
	 * {@link CommandLineParser#USER_ID_ARGUMENT_NAME},
	 * {@link CommandLineParser#SLEEP_TIME_ARGUMENT_NAME},
	 * {@link CommandLineParser#REPOSITORY_ARGUMENT_NAME},
	 * {@link CommandLineParser#LOG_FILE_ARGUMENT_NAME}
	 * {@link CommandLineParser#PREFERENCES_ARGUMENT_NAME}
	 * 
	 * However, the individual adapters might want to default a few arguments to make the launch command simpler
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		//load arguments into properties
		Properties arguments = new Properties();
		for (int i = 0; i < args.length; i = i + 2) {
			if ( args.length > i + 1) {
				arguments.put(args[i], args[i+1]);
			}
		}
		
		boolean firstStart = true;
		File configFile = null;
		if(arguments.containsKey(CommandLineParser.CONFIG_FILE_PATH)){//parameterExists(args, CommandLineParser.CONFIG_FILE_PATH)){
			String filePath = arguments.getProperty(CommandLineParser.CONFIG_FILE_PATH);//parameterValue(args, CommandLineParser.CONFIG_FILE_PATH);
			if(filePath != null){
				configFile = new File(filePath);
			}
			else
			{
				Logger.Log.error("Invalid arguments");
				System.exit(1);
			}
		}
		else{
			configFile = new File(CommandLineParser.CONFIG_FILE_NAME);
		}
		
		// First load the connection info object
		AdapterEncryption crypt = AdapterEncryption.getInstance();
		ConnectionInfo conn = CommandLineParser.loadArguments(args);
		
		if (conn == null) {
			Logger.Log.error("Invalid arguments");
			System.exit(1);
		}
		
		
		//Validate if adapter is running on RQM host
		try {
			String host = new URL(conn.getRepositoryAddress()).getHost();
			InetAddress addr = InetAddress.getLocalHost(); 
			String localAddr = addr.getHostAddress();
			if(host.equalsIgnoreCase("localhost")||
			   host.equalsIgnoreCase("loopback") ||		
			   host.equalsIgnoreCase("127.0.0.1")||
			   host.equalsIgnoreCase(localAddr))
			{
				Logger.Log.warn("Adapter is running applications on the RQM server, these applications could be used for nefarious activities. Running an adapter on the RQM server is not recommended. ");
			}
		} catch (MalformedURLException e2) {
			
			Logger.Log.warn("Error while validating host ");
		}
         catch (UnknownHostException e2) {
			
			Logger.Log.warn("Error while getting host address");
		}
		
         try {
        	 if (configFile != null && configFile.isDirectory()){
        		 throw new IllegalArgumentException(" -configFile accepts the path to a configuration file. The value passed is a directory.");
        	 }
        	 if ((configFile != null) && (configFile.exists()) ) {
        		 firstStart = false;
        		 Logger.Log.info("Adapter is reconnecting. Using saved configuration information from " + configFile.getName());
        	 }
        	 if(configFile != null && !configFile.exists()){
        		 Logger.Log.warn("Configuration file " + configFile.getAbsolutePath() +" does not exist, adapter will use the arguments passed");
        	 }
         }
         catch (Exception e) {
        	 Logger.Log.error("QMCommandLineAdapter : Error while starting");
        	 Logger.Log.error(e.getMessage());
        	 Logger.Log.debug(e);
        	 System.exit(1);
         }

         File prefsFile = null;
         Boolean foundSeleniumPrefsFile = false;
         try { // load preferences, use default file (Selenium or CLA) if explicit file not specified
        	 if (arguments.containsKey(CommandLineParser.PREFERENCES_ARGUMENT_NAME)) {
            	 prefsFile = new File((String)arguments.get(CommandLineParser.PREFERENCES_ARGUMENT_NAME));
        	 }
        	 else {
        		 File seleniumPrefs = new File(SELENIUM_PREFERENCES_NAME);
        		 if (seleniumPrefs.exists()) {
            		 prefsFile = seleniumPrefs;
        			 foundSeleniumPrefsFile = true;
        		 }
        		 else
            		 prefsFile = new File(DEFAULT_PREFERENCES_NAME);
        	 }
        	 new CommandLinePreferences(prefsFile);
         } catch ( Exception e ) {
        	 Logger.Log.warn("Preferences file can not be loaded : "+prefsFile);
        	 Logger.Log.debug(e);
         }

         // See if this adapter is configured to run Selenium/JUnit tests or as a standard command line adapter
		String adapterType = CLA_ADAPTER_TYPE;
		String adapterName = CLA_ADAPTER_DEFAULT_NAME;
		String adapterDesc = CLA_ADAPTER_DESCRIPTION;
 		String seleniumClasspath = CommandLinePreferences.prefs().getProperty(CommandLinePreferences.SELENIUM_CLASSPATH, null);

		if (seleniumClasspath != null && seleniumClasspath.length() > 0) {
			adapterType = SELENIUM_ADAPTER_TYPE;
			adapterName = SELENIUM_ADAPTER_DEFAULT_NAME;
			adapterDesc = SELENIUM_ADAPTER_DESCRIPTION;
		}
		else if (foundSeleniumPrefsFile) {
			/*
			 *  If the Selenium classpath property is not set correctly, and we are running the Selenium adapter,
			 *  we won't be able to run any Selenium tests, so log an error and exit.
			 */
       	 	Logger.Log.error(CommandLinePreferences.SELENIUM_CLASSPATH + " is not set correctly in " + SELENIUM_PREFERENCES_NAME + ", exiting!");
       	 	System.exit(1);
		}
		
		// Set a Java system property to specify the max file size for files attached to execution result via NewRequester.logAttachmentsFromFile()
		String maxAttachedFileSize = CommandLinePreferences.prefs().getProperty(AdapterConstants.RESULT_ATTACHMENT_FILE_PROPERTY_MAXSIZE, null);
		if (maxAttachedFileSize != null) 
			System.setProperty(AdapterConstants.RESULT_ATTACHMENT_FILE_PROPERTY_MAXSIZE, maxAttachedFileSize);
		
 		// Use Command line adapter's default name
		if (conn.getAdapterName().equals(CommandLineParser.DEFAULT_ADAPTER_NAME)) {
			conn.setAdapterName(adapterName);
		}
		
		String password = conn.getPassword();
		if (!firstStart && !(arguments.containsKey(CommandLineParser.PASSWORD_ARGUMENT_NAME)/*parameterExists(args, CommandLineParser.PASSWORD_ARGUMENT_NAME)*/)) {
			try {
				// check if the property isEncrypted is set. If so, decrypt else this was created earlier. Hence decode using base 64. But while saving to the file now, set the isEncrypted to true 
				// and encrypt before saving in store properties.
				if(AesEncryptUtil.isRQMErypted(password)) {
					password = crypt.decrypt(password);
				} else {
					password = new String((new Base64()).decode(password.getBytes("UTF-8")), "UTF-8"); // TODO change to decrypt - DD
				}
			} catch (UnsupportedEncodingException e1) {
				Logger.Log.error("QMCommandLineAdapter : password decode error");
				Logger.Log.debug(e1);
			}			
		}
		String proxyPassword = conn.getProxyPassword();
		if (!firstStart && !(arguments.containsKey(CommandLineParser.PROXY_PASSWORD_ARGUMENT_NAME))) {
			if(AesEncryptUtil.isRQMErypted(proxyPassword)) {
				proxyPassword = crypt.decrypt(proxyPassword);
			} 
		}
		
		ServerInformation server = new ServerInformation(conn.getRepositoryAddress());
		server.setLogin(conn.getUserId());
		server.setPassword(password);
		server.setProjectAreaAlias(conn.getProjectAreaAlias());
		
		//Set Proxy info under serverInformation instance
		server.setProxy(conn.getProxy());
		server.setProxyPort(conn.getProxyPort());
		server.setProxyUser(conn.getProxyUser());
		server.setProxyPassword(proxyPassword);
		AdapterInformation adapter = new AdapterInformation(conn.getAdapterId(), conn.getAdapterName(), adapterType);
		adapter.setDescription(adapterDesc);
		adapter.setPollingInterval((int)conn.getPollingInterval());
		adapter.setCapabilities(ADAPTER_CAPABILITIES);
		
		// Check to see if we are able to login to the RQM server.  If not, then exit.  This is confirming the
		// provided url,userId,password can login to RQM.
		NewRequester tempRequester = new NewRequester(server, null);
		// Convert may be project name or may be project alias name to project alias name
		String projectAlias = tempRequester.getProjectAreaAlias(conn.getProjectAreaAlias());
		if (projectAlias != null) {
			server.setProjectAreaAlias(projectAlias);
		}
		if (!tempRequester.wasInitialLoginSuccessful()) {
			Logger.Log
					.error("Failed to log in to server at "
							+ conn.getRepositoryAddress()
							+ ". Please check the repository URL "
							+ conn.getRepositoryAddress()
							+ ", user name "
							+ conn.getUserId()
							+ " and its password and try again. Make sure that the Server is available.");
			System.exit(1);
		}

		// Check if any problems encountered with ServerInformation parameters
		int serverStatus = tempRequester.getServerInfoStatus();
		if (serverStatus != NewRequester.SERVERINFO_VALID) {
			switch (serverStatus) {
			case NewRequester.SERVERINFO_NOT_PUBLIC_URI:
				String publicUri = tempRequester.getPublicURI();
				Logger.Log.error("The specified repository URL must use the public URI of the server: "	+ publicUri);
				System.exit(1);
			case NewRequester.SERVERINFO_UNKNOWN_PROJECTALIAS:
				String specifiedProjectArea = conn.getProjectAreaAlias() != null ? conn.getProjectAreaAlias() : server.getDefaultProjectAreaAlias();
				Map<String, String> aliasToNameMap = tempRequester.getAllProjectAreas();
				Iterator<String> aliasToNameKeyIterator = null;
				System.err.println("Attempting to connect using project area \""
								+ specifiedProjectArea
								+ "\" failed.  No such project area was found.");
				if (aliasToNameMap == null || aliasToNameMap.size() <= 0) {
					System.out.println("No project areas detected on the server.  Verify project area exist and specified user is able to login to the provided project area.");
				} else {
					System.out.println("Here is a list of Project Areas detected on server.  Please use -projectArea option to specify one of the following project area:");
					aliasToNameKeyIterator = aliasToNameMap.keySet().iterator();
					while (aliasToNameKeyIterator.hasNext()) {
						String alias = aliasToNameKeyIterator.next();
						String name = aliasToNameMap.get(alias);
						System.out.println("\tName: \"" + name + "\"\tAlias: " + alias);
					}
				}
				System.out.println("See -projectArea option in Command Line Adapter readme.txt file for more information.");
				System.exit(1);
			default:
				Logger.Log.error("Encountered an error while validating the specified server information.  Status code: " + serverStatus);
				System.exit(1);
			}
		}
				
		// This will login the adapter with the QM Server
		QMCommandLineAdapter commandLine = new QMCommandLineAdapter(server, adapter);
		
		// check if server registration exists
		boolean isRegistered = false;
		if (!firstStart){
			try {
				isRegistered = commandLine.getRequester().isAdapterRegistered();
				if (!isRegistered){
					Logger.Log.info("Stored configuration is out of sync. Adapter is re-registering.");
				}
			} catch (Exception e){
				isRegistered = false;
			}
		}
		
		// If its the first start, register and store properties into a config file
		Properties configProps = loadConfigFile(configFile);
		boolean regArgsChanged = isRegisteredArgsChanged(arguments, configProps);
		if(firstStart || !isRegistered){
			try {
				commandLine.getRequester().register();
			} catch (Exception e1) {
				Logger.Log.warn("Failed to save configuration file for the command line adapter");
				Logger.Log.debug(e1);
			}
		}
		if (firstStart || regArgsChanged || !AesEncryptUtil.isRQMErypted(configProps.getProperty(CommandLineParser.PASSWORD_PROPERTY_NAME)) || 
				(configProps.getProperty(CommandLineParser.PROXY_PASSWORD_PROPERTY_NAME) != null && !"".equals(configProps.getProperty(CommandLineParser.PROXY_PASSWORD_PROPERTY_NAME)) &&
				!AesEncryptUtil.isRQMErypted(configProps.getProperty(CommandLineParser.PROXY_PASSWORD_PROPERTY_NAME)))){			
			storeProperties(server, adapter, configFile, crypt);
		}
        
        Logger.Log.info("Created " + adapterName);
		Thread thread = commandLine.start();

		try {
			thread.join();
		} catch (InterruptedException exception) {
			// Ignore. The thread is done, that's all we care about.
		}
	}

	public static void storeProperties(ServerInformation server, AdapterInformation adapter, File configFile, AdapterEncryption crypt)
	{
		FileOutputStream fos = null;
		try{
			Properties properties = new Properties();
			properties.setProperty(CommandLineParser.REPOSITORY_PROPERTY_NAME, server.getUrl());
			properties.setProperty(CommandLineParser.ADAPTER_ID_PROPERTY_NAME, adapter.getId());
			properties.setProperty(CommandLineParser.USER_ID_PROPERTY_NAME, server.getLogin());
			properties.setProperty(CommandLineParser.PASSWORD_PROPERTY_NAME, crypt.encrypt(server.getPassword()));
			properties.setProperty(CommandLineParser.SLEEP_TIME_PROPERTY_NAME, Integer.toString(adapter.getPollingInterval()));
			properties.setProperty(CommandLineParser.ADAPTER_PROPERTY_NAME, adapter.getName());
			if(null != server.getProjectAreaAlias())//If default project area this will be null
				properties.setProperty(CommandLineParser.PROJECTAREA_ALIAS_PROPERTY_NAME, server.getProjectAreaAlias());
			
			//new paramters for proxy support
			if(null != server.getProxy())
				properties.setProperty(CommandLineParser.PROXY_PROPERTY_NAME, server.getProxy());
			if(null != server.getProxyPort())
				properties.setProperty(CommandLineParser.PROXY_PORT_PROPERTY_NAME, server.getProxyPort());
			if(null != server.getProxyUser())
				properties.setProperty(CommandLineParser.PROXY_USER_PROPERTY_NAME, server.getProxyUser());
			if(null != server.getProxyPassword())
				properties.setProperty(CommandLineParser.PROXY_PASSWORD_PROPERTY_NAME, crypt.encrypt(server.getProxyPassword()));
			
			fos = new FileOutputStream(configFile);
			properties.storeToXML(fos, COMMENT);
		} catch (FileNotFoundException e) {
			Logger.Log.error("storeProperties: Failed to save configuration file for the command line adapter");
			Logger.Log.debug(e);
		} catch (IOException e) {
			Logger.Log.error("storeProperties: Failed to save configuration file for the command line adapter");
			Logger.Log.debug(e);
		} finally {
			if (fos != null){
				try {
					fos.close();
				} catch (IOException e) {
					//do nothing
				}
			}
		}
	}
	
	private static boolean isRegisteredArgsChanged(Properties argsProps, Properties configFileProps)
	{
//		if(parameterExists(args, CommandLineParser.SLEEP_TIME_ARGUMENT_NAME) && !parameterValue(args, CommandLineParser.SLEEP_TIME_ARGUMENT_NAME).equals(propertyValue(CommandLineParser.SLEEP_TIME_PROPERTY_NAME, configFile))
//		|| parameterExists(args, CommandLineParser.ADAPTER_ID_ARGUMENT_NAME) && !parameterValue(args, CommandLineParser.ADAPTER_ID_ARGUMENT_NAME).equals(propertyValue(CommandLineParser.ADAPTER_ID_PROPERTY_NAME, configFile))
//		|| parameterExists(args, CommandLineParser.ADAPTER_ARGUMENT_NAME) && !parameterValue(args, CommandLineParser.ADAPTER_ARGUMENT_NAME).equals(propertyValue(CommandLineParser.ADAPTER_PROPERTY_NAME, configFile))
//		)
//			return true;
//		else
//			return false;
		if (		
				argsProps.containsKey(CommandLineParser.SLEEP_TIME_ARGUMENT_NAME) && !argsProps.get(CommandLineParser.SLEEP_TIME_ARGUMENT_NAME).equals(configFileProps.get(CommandLineParser.SLEEP_TIME_PROPERTY_NAME)) ||
				argsProps.containsKey(CommandLineParser.ADAPTER_ID_ARGUMENT_NAME) && !argsProps.get(CommandLineParser.ADAPTER_ID_ARGUMENT_NAME).equals(configFileProps.get(CommandLineParser.ADAPTER_ID_PROPERTY_NAME)) ||
				argsProps.containsKey(CommandLineParser.ADAPTER_ARGUMENT_NAME) && !argsProps.get(CommandLineParser.ADAPTER_ARGUMENT_NAME).equals(configFileProps.get(CommandLineParser.ADAPTER_PROPERTY_NAME))){
			return true;
		} else {
			return false;
		}
	}
	
	private static Properties loadConfigFile(File configFile){
		Properties props = new Properties();
		if (configFile == null || !configFile.exists()){
			return props;
   		}
		FileInputStream propertyFile = null;
		FileOutputStream outStream = null;
		try
		{
			propertyFile = new FileInputStream(configFile);
			if ( propertyFile != null )
			{
				props = new java.util.Properties() ;
				try {
					props.loadFromXML(propertyFile) ;
				} catch (InvalidPropertiesFormatException e) {
					propertyFile.close();
					propertyFile = new FileInputStream(configFile);
					props.load(propertyFile);
					// Resave to new xml utf-8 format
					outStream = new FileOutputStream(configFile);
					props.storeToXML(outStream, "Command Line Adapter Configuration");
				}
			}
		} catch(Throwable e){
			Logger.Log.error("Error loading configuration file properties.");
		} finally {
			if (propertyFile != null){
				try {
					propertyFile.close();
				} catch (IOException e) {
					//do nothing;
				}
			}
			if (outStream != null){
				try {
					outStream.close();
				} catch (IOException e) {
					//do nothing;
				}
			}
		}
		
		return props;
	}
}
